require 'rack'
require 'uu/os/util/class_scanner'
require 'uu/os/util/script_loader'
require 'uu/os/util/url_converter'
require 'uu/os/util/configuration_loader'
require 'uu/os/server/server_context'
require 'uu/os/server/rack_dispatcher'

module UU
  module OS
    module Server

      # Extension of {Rack::Builder} providing methods for easy initialization
      # and configuration of server without dependency on Rack API.
      module ServerInitializer
        include UU::OS::Util::ScriptLoader
        include UU::OS::Util::ClassScanner
        include UU::OS::Util::URLConverter
        include UU::OS::Util::ConfigurationLoader

        # @private
        @@boot_time = Time.now # :nodoc:

        # @private
        @@servlets = [] # :nodoc:

        # Initializes server. Typically invocation of this method will be the only
        # call within +config.ru+ configuration file. This method:
        #
        # * Adds source directories to Ruby load path.
        # * Loads configuration from +SERVER_CFG+ environment variable.
        # * Invokes all startup listeners (classes with defined static method +#on_server_startup+).
        # * Registers all controllers (internally realized via startup listener {UU::OS::Controller::ControllerLoader})
        # @param [Array<String>] directories First level directories within project structure where
        #    to look for startup listeners (Defaults to +lib+ and +app+ directory which are always loaded).
        # @return [void]
        # @note Order of startup listeners invocation is not guaranteed!
        def initialize_server(*directories)
          # TODO Is it safe also for production setup?
          # Does it only kills active thread or forces whole server to die?
          # We must test it.
          Thread.abort_on_exception = true if development_mode?
          # Prepare server context
          ctx = UU::OS::Server::ServerContext.new(self)
          # Add source directories to load path
          dirs = directories.dup
          dirs << 'lib' if ! dirs.include?('lib')
          dirs << 'app' if ! dirs.include?('app')
          dirs.each do |dir|
            $LOAD_PATH.unshift("./#{dir}")
          end
          # Load configuration
          hash_cls = Class.new(Hash)
          hash_cls.send(:include, UU::OS::Lang::HashConverter)
          @@configuration = hash_cls.new
          # Load standard configuration
          load_configuration(ENV['SERVER_CFG'])
          # Load optional configuration from embedded development server
          load_configuration(Thread.current[:SERVER_CFG])
          # Load all source scripts
          load_scripts(*dirs)
          # Find startup listeners
          listeners = find_classes do |clazz|
            clazz.respond_to?(:on_server_startup)
          end
          # Invoke startup listeners
          listeners.each do |listener|
            listener.on_server_startup(ctx)
          end
        end

        # Returns true if server is running in development/test mode.
        # @return [TrueClass, FalseClass] True if running in development/test mode.
        def development_mode?
          # When packed as war, RACK_ENV is automatically set to "production". In order
          # to be able to switch development mode when deployed on server, we have to
          # defile alternate environment variable (which is not overriden by rack).
          !!!(ENV['RACK_ENV'] =~ /^production$/i) || !!(ENV['SERVER_ENV'] =~ /^development$/i)
        end

        # Registers servlet on given path.
        # @param [Class<UU::OS::Server::AbstractServlet>] servlet_class Servlet class.
        # @param [TrueClass, FalseClass] ignore_context Flag if server (development) context path should
        #    be ignored and given servlet should be registered into root context (defaults to nil/false).
        # @return [void]
        # @param [Array<String>] path Path where to register servlet (path elements are concatenated).
        def register_servlet(servlet_class, ignore_context = nil, *path)
          if !ignore_context.nil? && !ignore_context.kind_of?(TrueClass) && !ignore_context.kind_of?(FalseClass)
            path.unshift(ignore_context)
            ignore_context = false
          end
          if development_mode? && !ignore_context
            development_context_path.each do |ctx_path|
              servlet_path = concat_path('/', ctx_path, *path)
              map servlet_path do
                run UU::OS::Server::RackDispatcher.new(self, servlet_class)
              end
              @@servlets << {path: servlet_path.to_s, :'class' => servlet_class.to_s }
            end
          else
            servlet_path = concat_path('/', *path)
            map servlet_path do
              run UU::OS::Server::RackDispatcher.new(self, servlet_class)
            end
            @@servlets << {path: servlet_path.to_s, :'class' => servlet_class.to_s }
          end
        end

        # @private
        alias_method :orig_load_configuration, :load_configuration # :nodoc:
        private :orig_load_configuration

        # Loads server configuration from given source. Method may be invoked
        # multiple times to load from multiple sources (configurations are
        # merged). Configuration should be always loaded before starting
        # server initialization with {#initialize} method.
        # @param [String, IO] source Configuration source. Accepted are:
        #    * Absolute or relative path to configuration file.
        #    * IO providing configuration content.
        #    * String with configuration content.
        # @return [void]
        def load_configuration(source)
          if source.kind_of?(Hash)
            @@configuration.merge!(source)
          else
            @@configuration.merge!(orig_load_configuration(source))
          end
        end

        # Returns server configuration.
        # @return [Hash<String, Object>] Server configuration.
        def configuration
          @@configuration
        end

        # Returns server development context path based on
        # mapping configuration files in +etc+ directory.
        # @return [Array<String>] Development context paths.
        def development_context_path
          @@dev_context ||= []
          if @@dev_context.empty?
            basename = File.basename(File.expand_path('.')).split('-')[0]
            context_path = basename.gsub('_', '-').downcase
            @@dev_context << context_path
          end
          @@dev_context
        end

        # Returns time when server was started.
        # @return [Time] Time when server was started.
        def boot_time
          @@boot_time
        end

        # Returns list of registered servlets.
        # @return [Array<Hash>] List of servlet registrations (each
        #   registration has +:path+ and +:class+ attributes).
        def registered_servlets
          @@servlets
        end

      end

    end
  end
end

# Extend {Rack::Builder}.
# @private
class Rack::Builder #:nodoc:
  include UU::OS::Server::ServerInitializer
end
